import os
import sys
from collections import defaultdict
from os import listdir
from os.path import (
    abspath,
    basename,
    dirname,
    isdir,
    join,
    split,
)

__version__ = '0.10.0'

from hammett.colors import (
    YELLOW,
    RED,
    RESET_COLOR,
    GREEN,
)

MISSING = object()

_orig_print = print


# pytest compatibility. I throw kwargs on the ground on purpose
def param(*args, **kwargs):
    return args


def parametrize(argnames, argvalues, indirect=False, ids=None, scope=None):
    def parametrize_wrapper(f):
        try:
            f.hammett_parametrize_stack.append((argnames, argvalues))
        except AttributeError:
            f.hammett_parametrize_stack = []
            f.hammett_parametrize_stack.append((argnames, argvalues))
        return f

    return parametrize_wrapper


class Item:
    pass


class _Mark:
    def __getattr__(self, name: str):
        if name == parametrize.__name__:
            return parametrize

        else:
            def marker(*args, **kwargs):
                args = list(args)

                def decorate(f):
                    try:
                        f.hammett_markers.append(Marker(name, args, kwargs))
                    except AttributeError:
                        f.hammett_markers = []
                        f.hammett_markers.append(Marker(name, args, kwargs))
                    return f

                if len(args) == 1:
                    if callable(args[0]):
                        f = args[0]
                        args[:] = args[1:]
                        return decorate(f)
                return decorate

            return marker


mark = _Mark()


class Marker:
    def __init__(self, name, args, kwargs):
        self.args = args
        self.kwargs = kwargs
        self.name = name

    def __repr__(self):
        return f'hammett.{self.name}'


class Globals:
    def __init__(self):
        self.source_location = None
        self.modules = None
        self.orig_cwd = os.getcwd()
        self.verbose = False
        self.durations = False
        self.terminal_width = None
        self.results = None
        self.settings = {}
        self.fail_fast = False
        self.quiet = False
        self.drop_into_debugger = False
        self.durations_results = []
        self.disable_assert_analyze = False
        self.output = []
        self.result_db = None
        self.should_stop = False
        self.hijacked_stdout = None
        self.hijacked_stderr = None
        self.use_cache = None
        self.pre_test_callback = None
        self.post_test_callback = None

    def reset(self):
        self.__init__()


g = Globals()


def print(arg='', end='\n', flush=False):
    g.output.append((arg, end, flush))
    if g.quiet:
        return
    _orig_print(arg, end=end, flush=flush, file=sys.__stdout__)


def fixture(*args, **kwargs):
    args = list(args)

    def decorator(f):
        from hammett.impl import register_fixture
        register_fixture(f, *args, **kwargs)
        return f

    if len(args) == 1:
        if callable(args[0]):
            f = args[0]
            args[:] = args[1:]
            return decorator(f)
        else:
            return decorator
    else:
        assert not args
        return decorator


# This is a deprecated feature in pytest, but we'll keep it here for now. pytest-django uses it for example.
def yield_fixture(*args, **kwargs):
    return fixture(*args, **kwargs)


def raises(expected_exception, match=None):
    from hammett.impl import RaisesContext
    return RaisesContext(expected_exception, match=match)


def warns(expected_warning=None, match=None):
    from hammett.impl import WarnsContext
    return WarnsContext(expected_warning, match=match)


def fail(message):
    raise RuntimeError(message)


class Option:
    def __getattr__(self, name):
        if hasattr(g, name):
            return getattr(g, name)
        return False


class Config:
    def __init__(self):
        self.option = Option()

    def getvalue(self, key):
        return g.settings['config'].get(key)

    def getini(self, name):
        return None

    # For xdist emulation add workerinput which is a dict that contains a key "workerid"


class PytestDirShim:
    def __init__(self, dir):
        self.dir = dir

    def dirpath(self):
        return self

    @property
    def purebasename(self):
        return basename(self.dir).rpartition('.')[0]

    def join(self, *args):
        return join(*([dirname(self.dir)] + [str(x) for x in args]))


class Request:
    current_fixture_setup = None

    def __init__(self, scope, parent, function=None):
        self.scope = scope
        self.function = function
        self.parent = parent
        self.additional_fixtures_wanted = set()
        self.keywords = {}
        self.fixturenames = set()
        self.finalizers = []
        self.fixture_results = {}
        self.funcargnames = []
        self.catch_log_handler = None
        if function:
            self.fspath = PytestDirShim(function.__code__.co_filename)
            self.name = function.__name__

        self.config = Config()

    def hammett_add_fixture_result(self, result):
        from hammett.impl import fixture_scope
        if fixture_scope.get(self.current_fixture_setup, 'function') != self.scope:
            self.parent.hammett_add_fixture_result(result)
        else:
            self.fixture_results[self.current_fixture_setup] = result

    def hammett_get_existing_result(self, name):
        try:
            return self.fixture_results[name]
        except KeyError:
            if self.parent:
                return self.parent.hammett_get_existing_result(name)
            return MISSING

    def teardown(self):
        for x in reversed(self.finalizers):
            x()

    @property
    def node(self):
        return self

    def addfinalizer(self, x):
        assert Request.current_fixture_setup is not None
        from hammett.impl import fixture_scope
        if fixture_scope.get(Request.current_fixture_setup, 'function') != self.scope:
            self.parent.addfinalizer(x)
        else:
            self.finalizers.append(x)

    def get_closest_marker(self, s):
        # TODO: fall back to the parent markers
        try:
            markers = self.function.hammett_markers
        except AttributeError:
            return None

        for marker in markers:
            if marker.name == s:
                return marker

        return None

    def getfixturevalue(self, s):
        return self.additional_fixtures_wanted.add(s)


def parse_markers(markers):
    if markers is None:
        return None

    def parse_one(s):
        if '[' not in s:
            return s, None

        assert s.endswith(']')
        base_name, _, name_plus_end_bracket = s.partition('[')
        return base_name, name_plus_end_bracket[:-1]

    # So far let's do something stupidly simple
    markers = markers.split(';')
    return dict(parse_one(m.strip()) for m in markers)


def guess_modules_and_source_path():
    if isdir('lib'):
        return listdir('lib'), 'lib'
    elif isdir('src'):
        return listdir('src'), 'src'

    this_dir = os.getcwd().split(os.sep)[-1]

    if this_dir.startswith('py-'):
        this_dir = this_dir[len('py-'):]

    if isdir(this_dir):
        return [this_dir], '.'
    elif isdir(this_dir.replace('-', '_')):
        return [this_dir.replace('-', '_')], '.'
    elif isdir(this_dir.replace(' ', '_')):
        return [this_dir.replace(' ', '_')], '.'
    elif isdir(this_dir.replace('-', '')):
        return [this_dir.replace('-', '')], '.'
    elif isdir(this_dir.replace(' ', '')):
        return [this_dir.replace(' ', '')], '.'
    return [], '.'


def is_test_file(f):
    return not f.startswith('.') and f.endswith('.py') and (f.endswith('__tests.py') or f.startswith('test_') or f.endswith('_test.py'))


def handle_dir(result, conftest_files, d):
    for root, dirs, files in os.walk(d):
        for f in files:
            if is_test_file(f):
                result.append(join(root, f) )
            if f == 'conftest.py':
                conftest_files.append(join(root, f))


def collect_files(filenames):
    result = []
    conftest_files = []
    if filenames is None:
        handle_dir(result, conftest_files, f'tests{os.sep}')
        handle_dir(result, conftest_files, f'test{os.sep}')
        handle_dir(result, conftest_files, f'docs{os.sep}')
        for module_name in g.modules:
            handle_dir(result, conftest_files, join(g.source_location, module_name))
    else:
        for filename in filenames:
            if os.path.exists(filename):
                if os.path.isdir(filename):
                    handle_dir(result, conftest_files, filename)
                else:
                    result.append(filename)
            else:
                # This is a symbol, module or function/class? Try module first.
                symbol_path = join(g.source_location, filename.replace('.', os.sep))
                if os.path.exists(symbol_path + '.py'):
                    result.append(symbol_path + '__tests.py')

    return result, conftest_files


def collect_file_data(path):
    data = {}
    for root, dirs, files in os.walk(path):
        dirs[:] = [x for x in dirs if not x.startswith('.') and x not in ['venv', 'env', '__pycache__']]
        for filename in files:
            if not filename.endswith('.py'):
                continue
            full_path = join(root, filename)
            if full_path.startswith(f'.{os.sep}'):
                full_path = full_path[2:]
            data[full_path] = os.stat(full_path).st_mtime_ns
    return data


DB_VERSION = 2
DB_FILENAME = '.hammett-db'


def write_result_db(results):
    """
    The results database is a simple pickled dict with some keys:

    db_version: the version. So we can change the format and throw away an old db if needed
    test_results: dict from filename -> (dict from test_name -> dict(stdout=str, stderr=str, status=str))
    file_data: dict filename -> nanosecond modification date
    """
    if not g.use_cache:
        return

    results['db_version'] = DB_VERSION
    from pickle import dump
    with open(DB_FILENAME, 'wb') as f:
        dump(results, f)


def new_result_db():
    return dict(
        db_version=DB_VERSION,
        test_results=defaultdict(dict),
        file_data=None,
    )


def read_result_db():
    if not g.use_cache:
        return new_result_db()
    from pickle import load
    try:
        with open(DB_FILENAME, 'rb') as f:
            results = load(f)
            if results['db_version'] != DB_VERSION:
                raise FileNotFoundError()
    except FileNotFoundError:
        return new_result_db()
    return results


def drop_cache_for_filename(result_db, filename):
    try:
        del result_db['test_results'][filename]
    except KeyError:
        pass


def update_result_db(result_db, new_file_data):
    if not g.use_cache:
        return

    if result_db['file_data'] is None:
        result_db['file_data'] = new_file_data
        assert not result_db['test_results']
        return result_db

    # Clear out test results when the test file or the tested module has changed
    old_file_data = result_db['file_data']
    clear_all_non_module_tests = False
    for filename, modification_time in old_file_data.items():
        if filename in new_file_data and modification_time != new_file_data[filename]:
            if filename.endswith('__tests.py') or filename.startswith('test_'):
                # The test has been changed
                drop_cache_for_filename(result_db, filename)
            else:
                # The module has been changed so translate the filename to the possible test files
                drop_cache_for_filename(result_db, filename[:-(len('.py'))] + '__tests.py')
                _, filename_only = split(filename)
                drop_cache_for_filename(result_db, f"tests{os.sep}{filename[:-(len('.py'))]}" + '__tests.py')
                clear_all_non_module_tests = True
                # TODO: this doesn't clear the db for module__function__tests.py

    if clear_all_non_module_tests:
        from pathlib import Path
        non_module_filenames = [x for x in result_db['test_results'].keys() if Path(x).stem.startswith('test_')]
        for x in non_module_filenames:
            del result_db['test_results'][x]

    result_db['file_data'] = new_file_data


def finish():
    for x in g.result_db['test_results'].values():
        for y in x.values():
            g.results[y.status] += 1


def main(verbose=False, fail_fast=False, quiet=False, filenames=None, drop_into_debugger=False, match=None, durations=False, markers=None, disable_assert_analyze=False, module_unload=False, cwd=None, use_cache=False, pre_test_callback=None, post_test_callback=None, capture_output=True, insert_cwd=True):
    params = main_setup(verbose=verbose, fail_fast=fail_fast, quiet=quiet, filenames=filenames, drop_into_debugger=drop_into_debugger, durations=durations, markers=markers, disable_assert_analyze=disable_assert_analyze, cwd=cwd, use_cache=use_cache, pre_test_callback=pre_test_callback, post_test_callback=post_test_callback, capture_output=capture_output, insert_cwd=insert_cwd)
    return main_run_tests(match=match, module_unload=module_unload, **params)


def main_setup(verbose=False, fail_fast=False, quiet=False, filenames=None, drop_into_debugger=False, durations=False, markers=None, disable_assert_analyze=False, cwd=None, use_cache=False, pre_test_callback=None, post_test_callback=None, capture_output=True, insert_cwd=True):
    import sys
    if sys.version_info[:2] < (3, 6):
        print('hammett requires python 3.6 or later')
        exit(999)

    sys.modules['pytest'] = sys.modules['hammett']

    markers = parse_markers(markers)

    clean_up_sys_path = False

    if cwd is None:
        cwd = os.getcwd()
    cwd = abspath(cwd)
    g.orig_cwd = cwd
    os.chdir(cwd)

    if cwd not in sys.path and insert_cwd:
        sys.path.insert(0, cwd)
        clean_up_sys_path = True

    g.reset()

    g.results = dict(success=0, failed=0, skipped=0, abort=0)
    g.verbose = verbose
    g.fail_fast = fail_fast
    g.quiet = quiet
    g.drop_into_debugger = drop_into_debugger
    g.durations = durations
    g.disable_assert_analyze = disable_assert_analyze
    g.use_cache = use_cache
    g.pre_test_callback = pre_test_callback
    g.post_test_callback = post_test_callback
    g.capture_output = capture_output

    from hammett.impl import read_settings
    read_settings()

    g.source_location = g.settings.get('source_location', '.')
    g.modules = g.settings.get('modules')
    if isinstance(g.modules, str):
        g.modules = [g.modules]
    if g.modules is None:
        m, s = guess_modules_and_source_path()
        g.source_location = g.source_location or s
        g.modules = m

    g.result_db = read_result_db()
    update_result_db(g.result_db, collect_file_data(g.source_location))
    write_result_db(g.result_db)

    filenames, conftest_files = collect_files(filenames)

    return dict(
        filenames=filenames,
        conftest_files=conftest_files,
        markers=markers,
        clean_up_sys_path=clean_up_sys_path,
    )


def main_run_tests(filenames, conftest_files, markers, clean_up_sys_path, match=None, module_unload=False, tests=None):
    if not filenames:
        print('No module to test found.')
        print('''You might need to add `modules` and `source_location` item under [hammett] in setup.cfg  like:

[hammett]
modules=
    foo
    bar
source_location=.
''')
        return 3

    g.results.update(dict(success=0, failed=0, skipped=0, abort=0))

    plugins_loaded = False
    conftests_loaded = False
    from os.path import split, sep

    session_request = Request(scope='session', parent=None)

    from hammett.impl import should_stop

    for test_filename in sorted(filenames):
        dirname, filename = split(test_filename)

        if dirname.startswith(f'.{os.sep}'):
            dirname = dirname[2:]

        module_name = f'{dirname.replace(sep, ".")}.{filename.replace(".py", "")}'
        # if module_name in sys.modules:
        #     del sys.modules[module_name]

        cache_filename = join(dirname, filename)
        if cache_filename in g.result_db['test_results']:
            from hammett.impl import print_status
            for test_name, result in g.result_db['test_results'][cache_filename].items():
                print_status(test_name, result)
            continue

        # We do this here because if all test results are up to date, we want to avoid loading slow plugins!
        if not plugins_loaded:
            from hammett.impl import load_plugins
            load_plugins()
            plugins_loaded = True

        if should_stop():
            break

        if not conftests_loaded:
            from hammett.impl import load_conftests
            load_conftests(conftest_files)
            conftests_loaded = True

        run_tests_for_filename(test_filename, session_request, markers, match, module_name, tests)

        # if module_unload:
        #     del sys.modules[module_name]

        if should_stop():
            break

    session_request.teardown()

    if not g.verbose:
        print()

    finish()

    color = GREEN
    if g.results['skipped']:
        color = YELLOW
    if g.results['failed']:
        color = RED

    if clean_up_sys_path:
        del sys.path[0]

    if g.durations:
        limit = 10
        print()
        print(f'--- {limit} slowest tests ---')
        g.durations_results.sort(key=lambda x: x[1], reverse=True)
        for name, time, _ in g.durations_results[:10]:
            print(f'{time}   {name}')

        print()
        print(f'--- {limit} slowest startup times for tests ---')
        print('Note: startup times might be shared cost so might show up unfairly')
        g.durations_results.sort(key=lambda x: x[2], reverse=True)
        for name, _, startup_time in g.durations_results[:10]:
            print(f'{startup_time}   {name}')

        print()

    print(f'{color}{g.results["success"]} succeeded, {g.results["failed"]} failed, {g.results["skipped"]} skipped{RESET_COLOR}')
    os.chdir(g.orig_cwd)

    if g.results['abort']:
        write_result_db(g.result_db)
        return 2

    write_result_db(g.result_db)
    return 1 if g.results['failed'] else 0


def load_module(module_name, test_filename):
    import importlib.util
    spec = importlib.util.spec_from_file_location(module_name, test_filename)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    try:
        spec.loader.exec_module(module)
    except Exception:
        print(f'Failed to load module {module_name}:')
        import traceback
        print(traceback.format_exc())
        g.results['abort'] += 1
    return module


def run_tests_for_filename(test_filename, session_request, markers, match, module_name, tests):
    module = load_module(module_name, test_filename)

    if g.results['abort']:
        return

    module_request = Request(scope='module', parent=session_request)

    module_markers = getattr(module, 'pytestmark', [])
    if not isinstance(module_markers, list):
        module_markers = [module_markers]

    from hammett.impl import execute_test_function, execute_test_class
    from unittest import TestCase
    from hammett.impl import should_stop
    for name, f in list(module.__dict__.items()):
        if g.results['abort']:
            break

        full_name = f'{test_filename}::{name}'.replace('./', '')

        is_test_function = name.startswith('test_') and callable(f)
        is_test_class = isinstance(f, type) and issubclass(f, TestCase) or name.startswith('Test')

        if not is_test_function and not is_test_class:
            continue

        if match is not None:
            if match not in name:
                continue

        if tests is not None:
            if full_name not in tests:
                continue

        for m in module_markers:
            f = m(f)

        if markers is not None:
            keep = False
            for f_marker in getattr(f, 'hammett_markers', []):
                if f_marker.name in markers:
                    arg = markers[f_marker.name]
                    if arg is not None:
                        assert len(f_marker.args) == 1, 'hammett only supports filtering on single arguments to markers right now'
                        if str(f_marker.args[0]) == arg:
                            keep = True
                            break
                    else:
                        keep = True
                        break
            if not keep:
                continue

        if is_test_function:
            execute_test_function(full_name, f, module_request)
        elif is_test_class:
            execute_test_class(full_name, f, module_request)

        if should_stop():
            break

    module_request.teardown()


def hookimpl(*_, **__):
    return lambda f: f


def multi_process_main(*, match, module_unload=False, **kwargs):
    import os

    # key == filename

    # manual fork
    result_by_key = {}

    def read_one_child_exit_status():
        pid, status = os.wait()
        result_by_key[key_from_pid[pid]] = (0xFF00 & status) >> 8  # The high byte contains the exit code

    hammett_kwargs = main_setup(**kwargs)

    key_from_pid = {}
    running_children = 0
    max_children = 160

    for key in hammett_kwargs['filenames']:
        pid = os.fork()
        if not pid:
            # In the child

            # this is needed for non-memory DBs
            Config.workerinput = dict(workerinput=f'_{key}')

            hammett_kwargs['filenames'] = [key]

            _orig_print('start:', key)

            result = main_run_tests(**hammett_kwargs, match=match, module_unload=module_unload)
            _orig_print('end:', key)
            if result != 0:
                # TODO: write failure information to stdout?
                pass
            os._exit(result)
        else:
            key_from_pid[pid] = key
            running_children += 1

        if running_children >= max_children:
            read_one_child_exit_status()
            running_children -= 1

    try:
        while True:
            read_one_child_exit_status()
    except ChildProcessError:
        pass
    return 0


def main_cli(args=None):
    if args is None:
        args = sys.argv[1:]
    from argparse import ArgumentParser
    parser = ArgumentParser(prog='hammett')
    parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', default=False)
    parser.add_argument('-x', dest='fail_fast', action='store_true', default=False)
    parser.add_argument('-q', dest='quiet', action='store_true', default=False)
    parser.add_argument('--multi-experimental', dest='multi_process', action='store_true', default=False)
    parser.add_argument('-k', dest='match', default=None)
    parser.add_argument('-m', dest='markers', default=None)
    parser.add_argument('-s', dest='capture_output', default=None)
    parser.add_argument('--use-cache', dest='use_cache', default=False, help='The cache is an experimental feature to run only relevant changes based on looking at what files have been changed.')
    parser.add_argument('--durations', dest='durations', action='store_true', default=False)
    parser.add_argument('--no-assert-analyze', dest='disable_assert_analyze', action='store_true', default=False)
    parser.add_argument('--pdb', dest='drop_into_debugger', action='store_true', default=False)
    parser.add_argument(dest='filenames', nargs='*')
    args = parser.parse_args(args)

    m = main
    if args.multi_process:
        m = multi_process_main

    return m(
        verbose=args.verbose,
        fail_fast=args.fail_fast,
        quiet=args.quiet,
        filenames=args.filenames or None,
        drop_into_debugger=args.drop_into_debugger,
        match=args.match,
        durations=args.durations,
        markers=args.markers,
        disable_assert_analyze=args.disable_assert_analyze,
        use_cache=args.use_cache,
        capture_output=args.capture_output,
    )


if __name__ == '__main__':
    exit(main_cli())



from collections import namedtuple


def tmpdir():
    from tempfile import TemporaryDirectory
    with TemporaryDirectory(prefix='hammet_') as d:
        yield d


_capsys_result = namedtuple('_capsys_result', ['out', 'err'])

# This is a bit over designed/weird because of pytest compatibility
def capsys():
    class CaptureFixture:
        def readouterr(self):
            import hammett
            return _capsys_result(hammett.g.hijacked_stdout.getvalue(), hammett.g.hijacked_stderr.getvalue())

    return CaptureFixture()


def caplog(request):
    """Access and control log capturing.

    Captured logs are available through the following properties/methods::

    * caplog.messages        -> list of format-interpolated log messages
    * caplog.text            -> string containing formatted log output
    * caplog.records         -> list of logging.LogRecord instances
    * caplog.record_tuples   -> list of (logger_name, level, message) tuples
    * caplog.clear()         -> clear captured records and formatted log output string
    """
    from hammett.fixtures_support import LogCaptureFixture
    from hammett.fixtures_support import catching_logs
    from hammett.fixtures_support import LogCaptureHandler
    import logging
    with catching_logs(LogCaptureHandler(), level=logging.INFO) as handler:
        request.catch_log_handler = handler
        result = LogCaptureFixture(request.node)
        yield result
        result._finalize()



# This file contents is copied whole sale from pytest. The license for pytest:

# The MIT License (MIT)
#
# Copyright (c) 2004-2020 Holger Krekel and others
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from io import StringIO
from typing import List
import logging
from contextlib import contextmanager
import re

_ANSI_ESCAPE_SEQ = re.compile(r"\x1b\[[\d;]+m")


def _remove_ansi_escape_sequences(text):
    return _ANSI_ESCAPE_SEQ.sub("", text)


@contextmanager
def catching_logs(handler, formatter=None, level=None):
    """Context manager that prepares the whole logging machinery properly."""
    root_logger = logging.getLogger()

    if formatter is not None:
        handler.setFormatter(formatter)
    if level is not None:
        handler.setLevel(level)

    # Adding the same handler twice would confuse logging system.
    # Just don't do that.
    add_new_handler = handler not in root_logger.handlers

    if add_new_handler:
        root_logger.addHandler(handler)
    if level is not None:
        orig_level = root_logger.level
        root_logger.setLevel(min(orig_level, level))
    try:
        yield handler
    finally:
        if level is not None:
            root_logger.setLevel(orig_level)
        if add_new_handler:
            root_logger.removeHandler(handler)


class LogCaptureHandler(logging.StreamHandler):
    """A logging handler that stores log records and the log text."""

    def __init__(self) -> None:
        """Creates a new log handler."""
        logging.StreamHandler.__init__(self, StringIO())
        self.records = []  # type: List[logging.LogRecord]

    def emit(self, record: logging.LogRecord) -> None:
        """Keep the log records in a list in addition to the log text."""
        self.records.append(record)
        logging.StreamHandler.emit(self, record)

    def reset(self) -> None:
        self.records = []
        self.stream = StringIO()


class LogCaptureFixture:
    """Provides access and control of log capturing."""

    def __init__(self, item) -> None:
        """Creates a new funcarg."""
        self._item = item
        # dict of log name -> log level
        self._initial_log_levels = {}  # type: Dict[str, int]

    def _finalize(self) -> None:
        """Finalizes the fixture.

        This restores the log levels changed by :meth:`set_level`.
        """
        # restore log levels
        for logger_name, level in self._initial_log_levels.items():
            logger = logging.getLogger(logger_name)
            logger.setLevel(level)

    @property
    def handler(self) -> LogCaptureHandler:
        """
        :rtype: LogCaptureHandler
        """
        return self._item.catch_log_handler  # type: ignore[no-any-return]  # noqa: F723

    def get_records(self, when: str) -> List[logging.LogRecord]:
        """
        Get the logging records for one of the possible test phases.

        :param str when:
            Which test phase to obtain the records from. Valid values are: "setup", "call" and "teardown".

        :rtype: List[logging.LogRecord]
        :return: the list of captured records at the given stage

        .. versionadded:: 3.4
        """
        handler = self._item.catch_log_handlers.get(when)
        if handler:
            return handler.records  # type: ignore[no-any-return]  # noqa: F723
        else:
            return []

    @property
    def text(self):
        """Returns the formatted log text."""
        return _remove_ansi_escape_sequences(self.handler.stream.getvalue())

    @property
    def records(self):
        """Returns the list of log records."""
        return self.handler.records

    @property
    def record_tuples(self):
        """Returns a list of a stripped down version of log records intended
        for use in assertion comparison.

        The format of the tuple is:

            (logger_name, log_level, message)
        """
        return [(r.name, r.levelno, r.getMessage()) for r in self.records]

    @property
    def messages(self):
        """Returns a list of format-interpolated log messages.

        Unlike 'records', which contains the format string and parameters for interpolation, log messages in this list
        are all interpolated.
        Unlike 'text', which contains the output from the handler, log messages in this list are unadorned with
        levels, timestamps, etc, making exact comparisons more reliable.

        Note that traceback or stack info (from :func:`logging.exception` or the `exc_info` or `stack_info` arguments
        to the logging functions) is not included, as this is added by the formatter in the handler.

        .. versionadded:: 3.7
        """
        return [r.getMessage() for r in self.records]

    def clear(self):
        """Reset the list of log records and the captured log text."""
        self.handler.reset()

    def set_level(self, level, logger=None):
        """Sets the level for capturing of logs. The level will be restored to its previous value at the end of
        the test.

        :param int level: the logger to level.
        :param str logger: the logger to update the level. If not given, the root logger level is updated.

        .. versionchanged:: 3.4
            The levels of the loggers changed by this function will be restored to their initial values at the
            end of the test.
        """
        logger_name = logger
        logger = logging.getLogger(logger_name)
        # save the original log-level to restore it during teardown
        self._initial_log_levels.setdefault(logger_name, logger.level)
        logger.setLevel(level)

    @contextmanager
    def at_level(self, level, logger=None):
        """Context manager that sets the level for capturing of logs. After the end of the 'with' statement the
        level is restored to its original value.

        :param int level: the logger to level.
        :param str logger: the logger to update the level. If not given, the root logger level is updated.
        """
        logger = logging.getLogger(logger)
        orig_level = logger.level
        logger.setLevel(level)
        try:
            yield
        finally:
            logger.setLevel(orig_level)



import importlib
import os
import sys
from collections.abc import Iterable
from warnings import catch_warnings

from dataclasses import dataclass
from typing import Optional
from unittest import SkipTest
from datetime import (
    datetime,
    timedelta,
)
from hammett import fixtures as built_in_fixtures

import hammett
from hammett.colors import (
    GREEN,
    RED,
    YELLOW,
    MAGENTA,
    RESET_COLOR,
)

_loading_plugins = False
_loading_conftest = False


class RaisesContext(object):
    def __init__(self, expected_exception, match):
        self.expected_exception = expected_exception
        self.excinfo = None
        self.match = match

    def __enter__(self):
        self.excinfo = ExceptionInfo()
        return self.excinfo

    def __exit__(self, *tp):
        __tracebackhide__ = True
        if tp[0] is None:
            assert False, f'Did not raise {self.expected_exception}'
        if self.match:
            import re
            assert re.match(self.match, str(tp[1]))
        self.excinfo.value = tp[1]
        self.excinfo.type = tp[0]
        suppress_exception = issubclass(self.excinfo.type, self.expected_exception)
        return suppress_exception


class ExceptionInfo:
    def __init__(self):
        self.type = None
        self.value = None

    def __str__(self):
        return f'<ExceptionInfo: type={self.type} value={self.value}'

    def __repr__(self):
        return str(self)


class WarnsContext(object):
    def __init__(self, expected_warning=None, match=None):
        self.expected_warning = expected_warning
        self.match_expr = match

    def __enter__(self):
        self.warning_catcher = catch_warnings(record=True)
        self.caught_warnings = self.warning_catcher.__enter__()
        return self.caught_warnings

    def __exit__(self, *tp):
        self.warning_catcher.__exit__()
        if tp[0]:
            return

        if self.expected_warning is not None:
            if not any(isinstance(w.message, self.expected_warning) for w in self.caught_warnings):
                found = '\n'.join(str(x) for x in self.caught_warnings)
                assert False, f'No warning of type {self.expected_warning} was emitted. These were found:\n\n{found}'

        if self.match_expr:
            import re
            if not any(re.search(self.match_expr, str(x.message)) for x in self.caught_warnings):
                found = '\n'.join(str(x) for x in self.caught_warnings)
                assert False, f'No warning matching {self.match_expr!r} was emitted. These were found:\n\n{found}'


class WarningInfo:
    def __init__(self):
        self.type = None
        self.value = None

    def __str__(self):
        return f'<WarningInfo: type={self.type} value={self.value}'

    def __repr__(self):
        return str(self)


fixtures = {}
auto_use_fixtures = set()
fixture_scope = {}


def fixture_function_name(f):
    r = f.__name__
    if r == '<lambda>':
        import inspect
        r = inspect.getsource(f)
    return r


# TODO: store args
def register_fixture(fixture, *args, autouse=False, scope='function', name=None, module_name=None):
    if scope == 'class':
        # hammett does not support class based tests
        return
    assert scope != 'package', 'Package scope is not supported at this time'

    if name is None:
        name = fixture_function_name(fixture)
    # pytest uses shadowing... I don't like it, but I guess we have to follow that?
    # assert name not in fixtures, 'A fixture with this name is already registered'
    if hammett.g.verbose and name in fixtures and name != 'request':
        hammett.print(f'{fixture} shadows {fixtures[name]}')
    if autouse:
        if _loading_plugins:
            module_name = ''
        else:
            module_name = module_name or fixture.__module__
        if _loading_conftest:
            module_name = module_name.replace('.conftest', '.')
        auto_use_fixtures.add((module_name, name))
    assert scope in ('function', 'class', 'module', 'package', 'session')
    fixture_scope[name] = scope
    fixtures[name] = fixture


for fixture_name, fixture in built_in_fixtures.__dict__.items():
    if not fixture_name.startswith('_'):
        register_fixture(fixture)


def pick_keys(kwargs, params):
    return {
        k: v
        for k, v in kwargs.items()
        if k in params
    }


_params_of_cache = {}


def params_of(f):
    if f not in _params_of_cache:
        import inspect
        _params_of_cache[f] = {
            x.name
            for x in inspect.signature(f).parameters.values()
            if x.kind in (inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY)
        }
    return _params_of_cache[f]


def _teardown_yield_fixture(fixturefunc, it):
    """Executes the teardown of a fixture function by advancing the iterator after the
    yield and ensure the iteration ends (if not it means there is more than one yield in the function)"""
    try:
        next(it)
    except StopIteration:
        pass
    else:
        hammett.print(f"yield_fixture {fixturefunc} function has more than one 'yield'")
        exit(1)


def call_fixture_func(fixturefunc, request, kwargs):
    # TODO: this fixture crashes for some reason, blacklist it for now
    if fixturefunc.__name__ == '_dj_autoclear_mailbox':
        return None
    existing_result = request.hammett_get_existing_result(fixture_function_name(fixturefunc))
    if existing_result is not hammett.MISSING:
        return existing_result

    import inspect
    import functools

    from hammett import Request
    Request.current_fixture_setup = fixture_function_name(fixturefunc)

    res = fixturefunc(**kwargs)

    yieldctx = inspect.isgenerator(res)
    if yieldctx:
        it = fixturefunc(**kwargs)
        res = next(it)
        finalizer = functools.partial(_teardown_yield_fixture, fixturefunc, it)
        request.addfinalizer(finalizer)

    request.hammett_add_fixture_result(res)

    Request.current_fixture_setup = None

    return res


def dependency_injection(f, fixtures, request):
    fixtures = fixtures.copy()

    relevant_auto_use_fixtures = {
        name
        for module_name, name in auto_use_fixtures
        if not module_name or f.__module__.startswith(module_name)
    }

    requested = params_of(f) | (request.additional_fixtures_wanted if request is not None else set()) | relevant_auto_use_fixtures

    request.additional_fixtures_wanted.clear()

    fully_resolved = dict(
        request=request,
    )

    unresolved = {name for name in requested if name not in fully_resolved}

    while unresolved:
        reduced = False
        for name in list(unresolved):
            if name not in fixtures:
                continue

            params = params_of(fixtures[name])

            # Add indirect dependencies to unresolved
            unresolved |= params

            # If we can resolve this dependency fully, remove it from unresolved list
            if params.issubset(set(fully_resolved.keys())):
                fully_resolved[name] = call_fixture_func(fixtures[name], request, pick_keys(fully_resolved, params))
                unresolved.remove(name)
                reduced = True

        if request.additional_fixtures_wanted:
            unresolved |= request.additional_fixtures_wanted
            request.additional_fixtures_wanted.clear()
            reduced = True

        if not reduced:
            break

    kwargs = {k: v for k, v in fully_resolved.items() if k in params_of(f)}
    return f, kwargs


def dependency_injection_and_execute(f, fixtures, orig_kwargs, request):
    f, kwargs = dependency_injection(f, fixtures, request)
    return f(**kwargs, **orig_kwargs)


def should_stop():
    return hammett.g.should_stop


def should_skip(_f):

    try:
        for marker in _f.hammett_markers:
            if marker.name == 'skip':
                return True
            if marker.name == 'skipif' and marker.args[0]:
                return True
    except AttributeError:
        pass

    return False


def indent(s, levels=1):
    ind = '    ' * levels
    return '\n'.join(ind + x for x in s.split('\n'))


def pretty_format(x, _indent=0):
    indent = '    ' * _indent
    result = ''
    if isinstance(x, dict):
        if not x:
            return '{}'
        result += '{\n'
        for k, v in x.items():
            result += f'{indent}    {k!r}: {pretty_format(v, _indent + 1)},\n'
        result += f'{indent}}}'
        return result
    elif isinstance(x, list):
        if not x:
            return '[]'
        result += '[\n'
        for v in x:
            result += f'{indent}    {pretty_format(v, _indent + 1)},\n'
        result += f'{indent}]'
        return result
    elif isinstance(x, tuple):
        if not x:
            return '(,)'
        result += '(\n'
        for v in x:
            result += f'{indent}    {pretty_format(v, _indent + 1)},\n'
        result += f'{indent})'
        return result
    else:
        return repr(x)


def build_feedback_for_exception():
    type, value, tb = sys.exc_info()
    while tb.tb_next:
        tb = tb.tb_next

    result = []

    local_variables = tb.tb_frame.f_locals
    if local_variables:
        result.append('--- Local variables ---')
        for k, v in sorted(local_variables.items()):
            result.append(f'{k}:')
            try:
                result.append(indent(pretty_format(v)))
            except Exception as e:
                result.append(f'   Error getting local variable repr: {e}')

    if type == AssertionError:
        result.append(analyze_assert(tb))

    return '\n'.join(result)


DIFF_STRING_SIZE_CUTOFF = 100


def analyze_assert(tb):
    if hammett.g.disable_assert_analyze:
        return ''

    result = []

    # grab assert source line
    try:
        with open(tb.tb_frame.f_code.co_filename) as f:
            source = f.read().split('\n')
    except FileNotFoundError:
        try:
            with open(os.path.join(hammett.g.orig_cwd, tb.tb_frame.f_code.co_filename)) as f:
                source = f.read().split('\n')
        except FileNotFoundError:
            result.append('Failed to analyze assert statement: file not found. Most likely there was a change of current directory.')
            return '\n'.join(result)

    line_no = tb.tb_frame.f_lineno - 1
    relevant_source = source[line_no]

    # if it spans multiple lines grab them all
    while line_no and not relevant_source.strip().startswith('assert '):
        line_no -= 1
        relevant_source = source[line_no] + '\n' + relevant_source

    if not relevant_source.strip().startswith('assert '):
        result.append('Failed to analyze assert statement (Did not find the assert)')
        return '\n'.join(result)

    import ast
    try:
        assert_statement = ast.parse(relevant_source.strip()).body[0]
    except SyntaxError:
        try:
            # grab one more line after the one where we got an exception, and try again
            relevant_source += '\n' + source[tb.tb_frame.f_lineno]
            assert_statement = ast.parse(relevant_source.strip()).body[0]
        except SyntaxError:
            result.append('Failed to analyze assert statement (SyntaxError)')
            return '\n'.join(result)

    # TODO: we should also analyze single values
    # We only analyze further if it's a comparison
    if assert_statement.test.__class__.__name__ != 'Compare':
        return '\n'.join(result)

    result.append('')
    result.append('--- Assert components ---')
    from astunparse import unparse
    try:
        left = eval(unparse(assert_statement.test.left), tb.tb_frame.f_globals, tb.tb_frame.f_locals)
        result.append('left:')
        result.append(indent(pretty_format(left)))
        right = eval(unparse(assert_statement.test.comparators), tb.tb_frame.f_globals, tb.tb_frame.f_locals)
    except Exception as e:
        result.append(f'Failed to analyze assert statement ({type(e)}: {e})')
        return '\n'.join(result)
    result.append('right:')
    result.append(indent(pretty_format(right)))
    if isinstance(left, str) and isinstance(right, str) and len(left) > DIFF_STRING_SIZE_CUTOFF and len(right) > DIFF_STRING_SIZE_CUTOFF and '\n' in left:
        result.append('')
        result.append('--- Diff of left and right assert components ---')
        left_lines = left.split('\n')
        right_lines = right.split('\n')
        from difflib import unified_diff
        for line in unified_diff(left_lines, right_lines, fromfile='expected', tofile='actual', lineterm=''):
            color = ''
            if line:
                color = {
                    '+': GREEN,
                    '-': RED,
                    '@': MAGENTA,
                }.get(line[0], '')
            result.append(f'{color}{line}{RESET_COLOR}')

    return '\n'.join(result)


SKIPPED = 'skipped'
SUCCESS = 'success'
FAILED = 'failed'


MESSAGES = {
    SKIPPED: dict(
        v=f' {YELLOW}Skipped{RESET_COLOR}',
        s=f'{YELLOW}s{RESET_COLOR}',
    ),
    SUCCESS: dict(
        v=f' {GREEN}Success{RESET_COLOR}',
        s=f'{GREEN}.{RESET_COLOR}',
    ),
    FAILED: dict(
        v=f' {RED}Failed{RESET_COLOR}',
        s=f'{RED}F{RESET_COLOR}',
    ),
}
MSG_SKIPPED_VERBOSE = f' {YELLOW}Skipped{RESET_COLOR}'
MSG_SKIPPED = f'{YELLOW}s{RESET_COLOR}'
MSG_SUCCESS_VERBOSE = f' {GREEN}Success{RESET_COLOR}'
MSG_SUCCESS = f'{GREEN}.{RESET_COLOR}'


@dataclass
class Result:
    status: str
    duration: Optional[timedelta] = 0
    stdout: str = ''
    stderr: str = ''
    stack_trace: str = ''
    feedback_for_exception: str = ''


def print_status(_name, result: Result):
    verbose = hammett.g.verbose
    message = MESSAGES[result.status]['v' if verbose else 's']
    if verbose:
        hammett.print(_name + ' ' + message + (str(result.duration) if result.duration else ''))
    else:
        hammett.print(message, end='', flush=True)

    if result.status == FAILED:
        hammett.print(RED)
        if not hammett.g.verbose:
            hammett.print()
        hammett.print('Failed: ', _name)
        hammett.print()

        hammett.print(result.stack_trace)

        hammett.print()
        if result.stdout:
            hammett.print(YELLOW)
            hammett.print('--- stdout ---')
            hammett.print(result.stdout)

        if result.stderr:
            hammett.print(RED)
            hammett.print('--- stderr ---')
            hammett.print(result.stderr)

        hammett.print(RESET_COLOR)

    if result.feedback_for_exception:
        hammett.print(result.feedback_for_exception)


def inc_test_result(_name, _f, result):
    print_status(_name, result)

    if hammett.g.fail_fast and result.status not in ('success', 'skipped'):
        hammett.g.should_stop = True

    filename = _f.__module__.replace('.', os.sep) + '.py'
    _name = _name[len(_f.__module__) + 1:]
    hammett.g.result_db['test_results'][filename][_name] = result


def run_test(_name, _f, _module_request, **kwargs):
    if should_skip(_f):
        inc_test_result(_name, _f, Result(status=SKIPPED))
        return

    if hammett.g.pre_test_callback is not None:
        hammett.g.pre_test_callback(name=_name, f=_f, module_request=_module_request, **kwargs)

    from io import StringIO

    req = hammett.Request(scope='function', parent=_module_request, function=_f)

    def request():
        return req

    register_fixture(request, module_name='', autouse=True)
    del request

    hammett.g.hijacked_stdout = StringIO()
    hammett.g.hijacked_stderr = StringIO()
    prev_stdout = sys.stdout
    prev_stderr = sys.stderr

    status = None
    duration = None
    start = None
    setup_time = None
    stack_trace = None
    feedback_for_exception = None

    if hammett.g.verbose:
        hammett.print(_name + '...', end='', flush=True)
    try:
        if hammett.g.capture_output:
            sys.stdout = hammett.g.hijacked_stdout
            sys.stderr = hammett.g.hijacked_stderr

        if hammett.g.durations:
            start = datetime.now()

        resolved_function, resolved_kwargs = dependency_injection(_f, fixtures, request=req)
        resolved_kwargs = {**resolved_kwargs, **kwargs}

        if hammett.g.durations:
            setup_time = datetime.now() - start
            start = datetime.now()

        resolved_function(**resolved_kwargs)

        if hammett.g.durations:
            duration = datetime.now() - start
            hammett.g.durations_results.append((_name, duration, setup_time))

        if hammett.g.capture_output:
            sys.stdout = prev_stdout
            sys.stderr = prev_stderr
        status = SUCCESS
    except KeyboardInterrupt:
        if hammett.g.capture_output:
            sys.stdout = prev_stdout
            sys.stderr = prev_stderr

        hammett.print()
        hammett.print('ABORTED')
        hammett.g.results['abort'] += 1
        hammett.g.should_stop = True
        return
    except SkipTest:
        if hammett.g.capture_output:
            sys.stdout = prev_stdout
            sys.stderr = prev_stderr

        status = SKIPPED
    except:
        if hammett.g.capture_output:
            sys.stdout = prev_stdout
            sys.stderr = prev_stderr

        import traceback
        stack_trace = traceback.format_exc()

        if not hammett.g.quiet:
            feedback_for_exception = build_feedback_for_exception()

        status = FAILED

    assert status is not None

    result = Result(
        status=status,
        duration=duration,
        stdout=hammett.g.hijacked_stdout.getvalue(),
        stderr=hammett.g.hijacked_stderr.getvalue(),
        stack_trace=stack_trace,
        feedback_for_exception=feedback_for_exception,
    )

    inc_test_result(_name, _f, result=result)

    if status == FAILED and hammett.g.drop_into_debugger:
        try:
            import ipdb as pdb
        except ImportError:
            import pdb
        pdb.set_trace()

    # Tests can change this which breaks everything. Reset!
    os.chdir(hammett.g.orig_cwd)
    req.teardown()

    if hammett.g.post_test_callback is not None:
        hammett.g.post_test_callback(_name=_name, _f=_f, _module_request=_module_request, **kwargs)



def execute_parametrize(_name, _f, _stack, _module_request, **kwargs):
    if not _stack:
        param_names = [f'{k}={v!r}' for k, v in kwargs.items()]
        _name = f'{_name}[{", ".join(param_names)}]'
        return run_test(_name, _f, _module_request=_module_request, **kwargs)

    names, param_list = _stack[0]
    if isinstance(names, str):
        names = [x.strip() for x in names.split(',')]
    # if you have just one argument, we allow to just pass a list instead of a list of lists, which is a pytest compatibility feature
    if len(names) == 1 and len(param_list) != 1 and not isinstance(param_list, Iterable):
        param_list = [param_list]
    for params in param_list:
        if not isinstance(params, (list, tuple)):
            params = [params]
        execute_parametrize(_name, _f, _stack[1:], _module_request, **{**dict(zip(names, params)), **kwargs})
        if should_stop():
            break


def execute_test_function(_name, _f, module_request):
    if getattr(_f, 'hammett_parametrize_stack', None):
        return execute_parametrize(_name, _f, _f.hammett_parametrize_stack, module_request)
    else:
        return run_test(_name, _f, module_request)


def execute_test_class(_name, _c, module_request):
    test_case = _c()
    try:
        test_case.setUp()
    except AttributeError:
        pass

    try:
        test_case.setup_class()
    except AttributeError:
        pass

    setup_method = getattr(test_case, 'setup_method', None)

    for member_name in dir(test_case):
        if member_name.startswith('test_'):
            test_func = getattr(test_case, member_name)
            if setup_method is not None:
                try:
                    setup_method(test_func)
                except Exception:
                    import traceback
                    print('Crash when running setup_method:')
                    traceback.print_exc()
                    continue
            execute_test_function(f'{_name}.{member_name}', test_func, module_request)

    try:
        test_case.tearDown()
    except AttributeError:
        pass


class FakePytestParser:
    def parse_known_args(self, *args):
        class Fake:
            pass
        s = Fake()
        s.itv = True
        s.ds = hammett.g.settings['django_settings_module']
        s.dc = None
        s.version = False
        s.help = False
        return s


class EarlyConfig:
    def __init__(self):
        self.inicfg = self
        self.config = self
        self.path = hammett.g.orig_cwd

    def addinivalue_line(self, name, doc):
        pass

    def getini(self, name):
        return None


early_config = EarlyConfig()


class OptionParserGroup:
    def addoption(self, *optnames, **attrs):
        pass

    def _addoption(self, *optnames, **attrs):
        pass


class OptionParser:
    def getgroup(self, name):
        return OptionParserGroup()

    def addini(self, *_, **__):
        pass


def load_plugin(module_name):
    import importlib
    try:
        plugin_module = importlib.import_module(module_name+'.plugin')
    except ImportError:
        plugin_module = importlib.import_module(module_name)

    parser = FakePytestParser()
    # noinspection PyBroadException
    try:
        if hasattr(plugin_module, 'pytest_load_initial_conftests'):
            plugin_module.pytest_load_initial_conftests(early_config=early_config, parser=parser, args=[])

        if hasattr(plugin_module, 'pytest_configure'):
            import inspect
            if inspect.signature(plugin_module.pytest_configure).parameters:
                plugin_module.pytest_configure(early_config)
            else:
                plugin_module.pytest_configure()

        if hasattr(plugin_module, 'pytest_addoption'):
            # This is just a stup for now
            plugin_module.pytest_addoption(OptionParser())
    except Exception:
        hammett.print(f'Loading plugin {module_name} failed: ')
        import traceback
        hammett.print(traceback.format_exc())
        hammett.g.results['abort'] += 1
        hammett.g.should_stop = True
        return
    try:
        importlib.import_module(module_name + '.fixtures')
    except ImportError:
        pass
    return True


def read_settings():
    from configparser import (
        ConfigParser,
        NoSectionError,
    )
    config_parser = ConfigParser()
    config_parser.read('setup.cfg')
    try:
        hammett.g.settings.update(dict(config_parser.items('hammett')))
        options = [x.partition('=') for x in hammett.g.settings.get('config', '').strip().split('\n')]
        options = [
            (x[0], x[2])
            for x in options
        ]
        hammett.g.settings['config'] = dict(options)
    except NoSectionError:
        return


def load_plugins():
    if 'plugins' not in hammett.g.settings:
        return

    global _loading_plugins
    _loading_plugins = True

    for plugin in hammett.g.settings['plugins'].strip().split('\n'):
        load_plugin(plugin)
        if should_stop():
            return

    _loading_plugins = False


def load_conftests(filenames):
    global _loading_conftest
    _loading_conftest = True

    load_conftest('conftest.py')
    for f in filenames:
        load_conftest(f)

    _loading_conftest = False


def load_conftest(filename):
    try:
        conftest = importlib.import_module(filename.replace('.py', '').replace(os.sep, '.'))
    except ImportError:
        return

    plugins = getattr(conftest, 'pytest_plugins', [])
    for plugin in plugins:
        load_plugin(plugin)
        if should_stop():
            return

    session_start = getattr(conftest, 'pytest_sessionstart', None)
    if session_start:
        session_start(None)
